%{
/*
----------------------------------------------------------------------------
                I N T E R M E D I A T E    F O R M 
----------------------------------------------------------------------
                Module  : if.lex.l
                Version : 1.4
                Date    : 05/22/98
                Authors : Marius
----------------------------------------------------------------------------
*/

#include "if2pml.h"
#include "y.tab.h"

extern YYSTYPE  yylval; 

extern char IfParserFilename[];                          /* input file */
extern char IfParserMessage[];                       /* message buffer */

/* char    yytext[512]; */ /* extern */
/* FILE    *yyin, *yyout; */

#define Token(y)    yylval = nn(y, yytext, ZS, ZN, ZN)


%}


SPACE                                   [ \t\n]
BLANC                                   {SPACE}+

DIGIT                                   [0-9]
NATURAL                                 {DIGIT}+
INTEGER                                 {NATURAL}
FLOAT                                   {INTEGER}"."{NATURAL}

LETTER                                  [a-zA-Z]
NAME                                    {LETTER}({LETTER}|{DIGIT}|_)*

COMMENT                                 "/*"([^*]|("*"[^/]))*"*/" 

STRING                                  \"[^"]*\"




%%

{BLANC}                                { /* nothing */ }
{COMMENT}                              { /* nothing */ }


abstract                               { Token(ABSTRACT); return ABSTRACT; }
active                                 { Token(ACTIVE); return ACTIVE; }
and                                    { Token(AND); return AND; }
any                                    { Token(ANY); return ANY; }
array                                  { Token(ARRAY); return ARRAY; }
bag                                    { Token(BAG); return BAG; }
bool                                   { Token(BOOL); return BOOL; }
buffer                                 { Token(BUFFER); return BUFFER; }
clock                                  { Token(CLOCK); return CLOCK; }
discard                                { Token(DISCARD); return DISCARD; }
do                                     { Token(DO); return DO; }
end                                    { Token(END); return END; }
enum                                   { Token(ENUM); return ENUM; }
from                                   { Token(FROM); return FROM; }
gate                                   { Token(GATE); return GATE; }
if                                     { Token(IF); return IF; }
implementation                         { Token(IMPLEMENTATION); return IMPLEMENTATION; }
in                                     { Token(IN); return IN; }
input                                  { Token(INPUT); return INPUT; }
int                                    { Token(INT); return INT; }
nat                                    { Token(NAT); return NAT; }
not                                    { Token(NOT); return NOT; }
of                                     { Token(OF); return OF; }
or                                     { Token(OR); return OR; }
output                                 { Token(OUTPUT); return OUTPUT; }
pid                                    { Token(PID); return PID; }
pos                                    { Token(POS); return POS; }
pred                                   { Token(PRED); return PRED; }
process                                { Token(PROCESS); return PROCESS; }
queue                                  { Token(QUEUE); return QUEUE; }
range                                  { Token(RANGE); return RANGE; }
real                                   { Token(REAL); return REAL; }
record                                 { Token(RECORD); return RECORD; }
reset                                  { Token(RESET); return RESET; }
tpc                                    { Token(TPC); return TPC; }
timer                                  { Token(TIMER); return TIMER; }
to                                     { Token(TO); return TO; }
transition                             { Token(TRANSITION); return TRANSITION; }
type                                   { Token(TYPE); return TYPE; }
save                                   { Token(SAVE); return SAVE; }
set                                    { Token(SET); return SET; }
signal                                 { Token(SIGNAL); return SIGNAL; }
stack                                  { Token(STACK); return STACK; }
state                                  { Token(STATE); return STATE; }
succ                                   { Token(SUCC); return SUCC; }
sync                                   { Token(SYNC); return SYNC; }
system                                 { Token(SYSTEM); return SYSTEM; }
var                                    { Token(VAR); return VAR; }




".."                                   { Token(DOTS); return DOTS; }
"|||"                                  { Token(NSYNC); return NSYNC; }
"||"                                   { Token(FSYNC); return FSYNC; }
"|["                                   { Token(LSYNC); return LSYNC; }
"]|"                                   { Token(RSYNC); return RSYNC; }
":="                                   { Token(ASGN); return ASGN; }
"<="                                   { Token(LE); return LE; }
">="                                   { Token(GE); return GE; }
"<>"                                   { Token(NE); return NE; }



"("                                    { Token('('); return '('; }
")"                                    { Token(')'); return ')'; }
"["                                    { Token('['); return '['; }
"]"                                    { Token(']'); return ']'; }
"{"                                    { Token('{'); return '{'; }
"}"                                    { Token('}'); return '}'; }
"+"                                    { Token('+'); return '+'; }
"-"                                    { Token('-'); return '-'; }
"*"                                    { Token('*'); return '*'; }
"/"                                    { Token('/'); return '/'; }
"%"                                    { Token('%'); return '%'; }
"<"                                    { Token('<'); return '<'; }
">"                                    { Token('>'); return '>'; }
"="                                    { Token('='); return '='; }
"."                                    { Token('.'); return '.'; }
":"                                    { Token(':'); return ':'; }
";"                                    { Token(';'); return ';'; }
","                                    { Token(','); return ','; }
"?"                                    { Token('?'); return '?'; }
"!"                                    { Token('!'); return '!'; }
"|"                                    { Token('|'); return '|'; }


{NAME}                                 { Token(NAME); return NAME; }
{STRING}                               { Token(STRING); return STRING; }
{FLOAT}                                { Token(FLOAT); return FLOAT; }
{INTEGER}                              { Token(INTEGER); return INTEGER; }


. {
  sprintf(IfParserMessage,
	  "%s(%d): lexical error (at or near '%s')", 
	  IfParserFilename, yylineno, yytext); 
  IF_ERROR(IfParserMessage);
  }



%% 

/*
----------------------------------------------------------------------------
Lex Specific Functions
----------------------------------------------------------------------------
*/

int yywrap()
{ 
                         /* yywrap implementation */
   return 1; 
}
