%{
/*
----------------------------------------------------------------------------
                I N T E R M E D I A T E    F O R M 
----------------------------------------------------------------------------
                Module  : if.yacc.y
                Version : 1.14
                Date    : 05/26/98
                Authors : Marius
----------------------------------------------------------------------------
*/

#include <stdio.h>
#include "if2pml.h"

/* extern FILE *yyin, *yyout; */

int yylineno;                                       /* scanner line */
char yytext[512];                                    /* scanner token */

/* static char IfParserFilename[256]; */           /* input file */
/* static char IfParserMessage[256];  */              /* message buffer */


inline void yyerror(char* msg){                  /* syntax error handling */                
  sprintf(IfParserMessage,
	  "%s(%d): %s (at or near '%s')",
	  IfParserFilename, yylineno, msg, yytext);
  IF_ERROR(IfParserMessage);
}

#define NYI yyerror("not yet implemented");

SyStruct *systr;
SigStruct *sigstr;
VarStruct *vstr; 
BuffStruct *buffstr;
ProcStruct *prstr;
StateStruct *ststr;
FilterActionStruct *fastr;
TransStruct *trstr;
InputStruct *inpstr;

Node *syst; /* pointer to the SYSTEM type node */

%}


/*
----------------------------------------------------------------------------
Tokens
----------------------------------------------------------------------------
*/

%token SYSTEM


%token TYPE

%token INT REAL BOOL PID NAT CLOCK TIMER

%token ENUM RANGE RECORD ARRAY ABSTRACT 

%token DOTS OF IMPLEMENTATION END


%token SIGNAL GATE


%token VAR

%token BUFFER QUEUE BAG SET STACK 


%token SYNC


%token PROCESS 

%token STATE

%token IF DISCARD SAVE IN TPC

%token TRANSITION

%token FROM INPUT DO ASGN OUTPUT RESET TO


%token INTEGER FLOAT

%token ANY /* SENDER SELF */


%token NAME STRING




/*
----------------------------------------------------------------------------
Operator Precedence
----------------------------------------------------------------------------
*/

%left NSYNC FSYNC LSYNC RSYNC

%left '|'

%left ','

%right '?' ':'

%right '!'

%left OR 

%left AND

%left '=' NE

%left '>' '<' GE LE

%left '+' '-'

%left '*' '/' '%'

%right NOT UMIN PRED SUCC POS ACTIVE

%left '.' '[' '('

%%




/*
----------------------------------------------------------------------------
System Definition Syntax
----------------------------------------------------------------------------
*/

SystemDefinition: 
          SYSTEM Name ';' 
          TypeSection 
          SignalSection 
          GateSection
          VariableSection 
          BufferSection 
          SynchronisationSection
          ProcessDefinitionSet
                                   {
                                   systr = (SyStruct *)
                                          emalloc(sizeof(SyStruct));
				   systr->types = $4;
				   systr->signals = $5;
				   systr->buffers = $8;
				   systr->processes = $10;
                                   syst = nn(SYSTEM, $2->txt, systr, ZN, ZN);
                                   }
        ;

TypeSection: 
          TYPE TypeDefinitionSet                   { $$ = $2; }
        | /* empty */                              { $$ = ZN; }
        ;

TypeDefinitionSet: 
          TypeDefinitionSet TypeDefinition         { $$ = insertl($1,$2); }
        | TypeDefinition                           { $$ = $1; }
        ;

SignalSection: 
          SIGNAL SignalDefinitionSet               { $$ = $2; }
        | /* empty */                              { $$ = ZN; }
        ;

SignalDefinitionSet:    
          SignalDefinitionSet SignalDefinition    { $$ = insertl($1,$2); }
        | SignalDefinition                        { $$ = $1; }
        ;

GateSection: 
          GATE GateDefinitionSet                   { }
        | /* empty */                              { }
        ;

GateDefinitionSet: 
          GateDefinitionSet GateDefinition         { }
        | GateDefinition                           { }
        ;

VariableSection: 
          VAR VariableDefinitionSet                { $$ = $2; }
        | /* empty */                              { $$ = ZN; }
        ;

VariableDefinitionSet: 
          VariableDefinitionSet VariableDefinition { $$ = insertl($1,$2); }
        | VariableDefinition                       { $$ = $1; }
        ;

BufferSection: 
          BUFFER BufferDefinitionSet               { $$ = $2; }
        | /* empty */                              { $$ = ZN; }
        ;

BufferDefinitionSet: 
          BufferDefinitionSet BufferDefinition     { $$ = insertl($1,$2); }
        | BufferDefinition                         { $$ = $1;}
        ;

SynchronisationSection: 
          SYNC Synchronisation END ';'             { }
        | /* empty */                              { }
        ;


ProcessDefinitionSet: 
          ProcessDefinitionSet ProcessDefinition   { $$ = insertl($1,$2); }
        | ProcessDefinition                        { $$ = $1;}
        ;




/*
----------------------------------------------------------------------------
Type Definition Syntax
----------------------------------------------------------------------------
*/

TypeDefinition: 
          Name '=' ENUM ItemDefinitionSet ';'      
                                  { $$ = nn(ENUM, $1->txt, ZS, $4, ZN); }
        | Name '=' RANGE RangeDefinition ';'       
                                  { $$ = nn(RANGE, $1->txt, ZS, $4, ZN); }
        | Name '=' RECORD  VariableDefinitionSet END ';'         
                                  { $$ = nn(RECORD, $1->txt, ZS, $4, ZN); }
        | Name '=' ARRAY '[' RangeDefinition ']' OF TypeName ';' 
                                  { $$ = nn(ARRAY, $1->txt, ZS,  
					nn(ARRAY, ZC, ZS, $5, $8), ZN); }
        | Name '=' ABSTRACT FunctionDefinitionOptSet ImplementationSection END ';' 
                                  { $$ = nn(ABSTRACT, $1->txt, ZS, $1, ZN); }
        ;

ItemDefinitionSet: 
          ItemDefinitionSet ',' ItemDefinition     { $$ = insertl($1,$3); }
        | ItemDefinition                           { $$ = $1; }
        ;

ItemDefinition: 
         Name                                      { $$ = $1; }
        ;

RangeDefinition: 
          Integer DOTS Integer                     
                                            { $$ = nn(DOTS, ZC, ZS, $1, $3); }
        | Integer DOTS '-' Integer                
                                      { $$ = nn('-', ZC, ZS, $1, $4); }
                      /* negative indices are forbidden for arrays " */
        | '-' Integer DOTS Integer       
                                      { $$ = nn('-', ZC, ZS, $2, $4); }
        | '-' Integer DOTS '-' Integer            
                                      { $$ = nn('-', ZC, ZS, $2, $5); }
        | Name DOTS Name                 
                                      { $$ = nn('-', ZC, ZS, $1, $3); } 
        ;


FunctionDefinitionOptSet:
          FunctionDefinitionOptSet FunctionDefinition            { }
        | /* empty */                                            { }
        ;

FunctionDefinition: 
          Name '(' TypeArgOptList ')' ':' TypeName ';'           { }
        ;


ImplementationSection: 
          IMPLEMENTATION String                    { }
        | /* empty */                              { }
        ;

TypeArgOptList: 
          TypeArgList                              { }
        | /* empty */                              { }
        ;

TypeArgList: 
	  TypeArgList ',' TypeName                 { $$ = insertl($1, $3);}
        | TypeName                                 { $$ = $1; }
        ;

TypeName: 
	  Name                                     { $$ = $1; }
        | BOOL                                     { $$ = $1; }
        | INT                                      { $$ = $1; }
        | REAL                                     { $$ = $1; }
        | NAT                                      { $$ = $1; }
        | CLOCK                                    { $$ = $1; }
        | TIMER                                    { $$ = $1; }
        | PID                                      { $$ = $1; }
        ;



/*
----------------------------------------------------------------------------
Signal Definition Syntax
----------------------------------------------------------------------------
*/

SignalDefinition: 
          Name ';'                                 
                               {
				 sigstr = (SigStruct *)
				   emalloc(sizeof(SigStruct));
				 $1->nstr = sigstr;
				 $$ = $1;
			       }
        | Name '(' TypeArgList ')' ';'             
                               {
				 sigstr = (SigStruct *)
				   emalloc(sizeof(SigStruct));
				 $1->nstr = sigstr;
				 $1->lft = $3;
			         $$ = $1;
			       }
        ;

SignalArgOptList: 
	  '*'                                      { $$ = $1; }
        | SignalArgList                            { $$ = $1; }
        ;

SignalArgList:
	  SignalArgList ',' SignalName             { $3->rgt = $1; $$ = $3; }
        | SignalName                               { $$ = $1; }
        ;

SignalName: 
          Name  { $$ = $1; }
        ;



/*
----------------------------------------------------------------------------
Gate Definition Syntax
----------------------------------------------------------------------------
*/

GateDefinition: 
          Name ';'                                 { }
        | Name '(' TypeArgList ')' ';'             { }
        ;


GateArgList:
	  GateArgList ',' GateName                 { }
	| GateName                                 { }
        ;

GateName: 
          Name                                     { }
        ;



/*
----------------------------------------------------------------------------
Variable Definition Syntax
----------------------------------------------------------------------------
*/

VariableDefinition: 
          Name Default ':' TypeName ';'            
                       {
			 vstr =
                            (VarStruct *) emalloc(sizeof(VarStruct));
			 vstr->vtyp = $4;
			 vstr->dflt = $2;
			 $$ = nn(VAR, $1->txt, vstr, ZN, ZN);
                       }
        | Name Default ',' VariableDefinition      
                       {
			 vstr =
                            (VarStruct *) emalloc(sizeof(VarStruct));
			 vstr->vtyp = ZN;
			 vstr->dflt = $2;
			 $$ = nn(VAR, $1->txt, vstr, $4, ZN);
                        }
        ;

Default: 
          '(' ExpressionArgList ')'                { $$ = $2; }
        | /* empty */                              { $$ = ZN; } 
        ;




/*
----------------------------------------------------------------------------
Buffer Definition Syntax
----------------------------------------------------------------------------
*/

BufferDefinition: 
          Name ':' BufferKind OptionSet OF SignalArgOptList ';'  
                    {
                     buffstr =
                            (BuffStruct *) emalloc(sizeof(BuffStruct));
                     buffstr->signals = $6;
                     $$ = nn(0, $1->txt, buffstr, ZN, ZN); 
                     }
        ;

BufferKind:               
          QUEUE                                    { }
        | BAG                                      { }
        | SET                                      { }
        | STACK                                    { }
        ;

BufferName: 
          Name                                     { $$=$1; } /* !? */
        ;



/*
----------------------------------------------------------------------------
Synchronisation Expression Syntax
----------------------------------------------------------------------------
*/

Synchronisation: 
          Name                                     { }
        | Synchronisation NSYNC Synchronisation    { }
        | Synchronisation FSYNC Synchronisation    { }
        | Synchronisation LSYNC GateArgList RSYNC Synchronisation  { }
        | '(' Synchronisation ')'                  { }
        ;



/*
----------------------------------------------------------------------------
Process Definition Syntax
----------------------------------------------------------------------------
*/

ProcessDefinition:  
          PROCESS Name /* Instances */ OptBuffer ';' 
          VariableSection
          STATE
          StateDefinitionSet
          TRANSITION 
          TransitionDefinitionSet
                         {
                          prstr = 
                             (ProcStruct *) emalloc(sizeof(ProcStruct)); 
                          prstr->buffer = $3;
			  prstr->vars = $5;
			  prstr->states = $7;
			  prstr->trans = $9;
                          $$ = nn(PROCESS, $2->txt, prstr, ZN, ZN);
                          }
        ;

Instances: 

          '(' Integer ')'                          { $$ = $2; }
        | /* empty */                              { $$ = ZN; }
        ;

StateDefinitionSet: 
          StateDefinitionSet StateDefinition       { $$ = insertl($1, $2); }
        | StateDefinition                          { $$ = $1; }
        ;

TransitionDefinitionSet: 
          TransitionDefinitionSet TransitionDefinition 
                                                   { $$ = insertl($1,$2); }
        | TransitionDefinition                     { $$ = $1; }
        ;

OptBuffer:
          ':' BUFFER BufferName              { $$ = $3; }
        | /* empty */                        { $$ = ZN; }
        ;


/*
----------------------------------------------------------------------------
State Definition Syntax
----------------------------------------------------------------------------
*/

StateDefinition: 
          Name OptionSet ';'                       
                    {
                     ststr =
                            (StateStruct *) emalloc(sizeof(StateStruct));
		     ststr->options = $2;
		     $$ = nn(STATE, $1->txt, ststr, ZN, ZN);
                     }
        | Name OptionSet 
          DiscardSection 
          SaveSection 
          TpcSection
          END ';'
 
                    {
		     ststr =
                            (StateStruct *) emalloc(sizeof(StateStruct));
		     ststr->options = $2;
                     ststr->discards = $3;
		     ststr->saves = $4; /* !? */
		     $$ = nn(STATE, $1->txt, ststr, ZN, ZN); 
		    }
        ;

DiscardSection: 
          DISCARD FilterActionSet                  { $$ = $2; }
        | /* empty */                              { $$ = ZN; }
        ;

SaveSection: 
          SAVE FilterActionSet                     { $$ = $2; } /* !? */
        | /* empty */                              { $$ = ZN; } /* !? */
        ;

FilterActionSet: 
          FilterActionSet FilterAction             { $$ = insertl($1,$2); }
        | FilterAction                             { $$ = $1; }
        ;

FilterAction: 
	  SignalArgOptList IN BufferName OptGuard ';'          
                             { 
			      fastr =
				 (FilterActionStruct *) 
                                  emalloc(sizeof(FilterActionStruct));
			      fastr->signals = $1;
			      fastr->buffer = $3;
			      fastr->guard = $4;
			      $$ = nn(DISCARD, ZC, fastr, ZN, ZN);
                                }
        | SignalArgOptList OptGuard ';'            
                            {
			      fastr =
				 (FilterActionStruct *) 
                                  emalloc(sizeof(FilterActionStruct));
			      fastr->signals = $1;
			      fastr->guard = $2;
			      $$ = nn(DISCARD, ZC, fastr, ZN, ZN);
			    }
        ;

TpcSection: 
          TPC Expression ';'                       { }
        | /* empty */                              { }
        ;

OptGuard: 
          IF Expression                            { $$ = $2; }
        | /* empty */                              { $$ = ZN; }
        ;

StateName: 
          Name                                     { $$ = $1; }
        ;




/*
----------------------------------------------------------------------------
Transition Definition Syntax
----------------------------------------------------------------------------
*/

TransitionDefinition: 
          FROM SourceState OptionSet 
          OptGuard 
	  InputSection 
	  BodySection 
	  TO TargetState ';'
                           {
			    trstr =
			      (TransStruct *) emalloc(sizeof(TransStruct));
			    trstr->source = $2;
			    trstr->optionset = $3;
			    trstr->guard = $4;
			    trstr->input = $5;
			    trstr->body = $6;
			    trstr->target = $8;
			    $$=nn(TRANSITION, ZC, trstr, ZN, ZN); 
			   }
        ;

SourceState: 
          StateName                                { $$ = $1; }
        | '*'                                      { $$ = $1; } /* !? */
        ;

TargetState: 
	  StateName                                { $$ = $1; }
        | '-'                                      { $$ = $1; } /* !? */
        ;

InputSection: 
          INPUT SignalName FROM BufferName OptGuard                            {
inpstr =
      (InputStruct *) emalloc(sizeof(InputStruct));
      inpstr->signal = $2;
      inpstr->frombuff = $4;
      inpstr->guard = $5;
      $$ = nn(INPUT, ZC, inpstr, ZN, ZN);	    
 }
        | INPUT SignalName '(' ExpressionArgList ')' FROM BufferName OptGuard  {
      inpstr =
          (InputStruct *) emalloc(sizeof(InputStruct));
      inpstr->signal = $2;
      inpstr->sigexpr = $4;
      inpstr->frombuff = $7;
      inpstr->guard = $8;
      $$ = nn (INPUT, ZC, inpstr, ZN, ZN);
} /* !? */
        | SYNC GateName OffertOptSet OptGuard   
{  inpstr =
      (InputStruct *) emalloc(sizeof(InputStruct));
      inpstr->signal = $2; /* gate */
      inpstr->frombuff = $3; /* offert */
      inpstr->guard = $4;
      $$ = nn(SYNC, ZC, inpstr, ZN, ZN); 
}
/* !? */
        | /* empty */                              { $$ = ZN; }
        ;

BodySection: 
          DO BodyAction                            { $$ = $2; }
        | /* empty */                              { $$ = ZN; }
        ;

BodyAction: 
          BodyAction '|' BodyAction                
                      { $$ = nn('|', ZC, ZS, $1, $3); }    /* !? */
        | BodyAction ',' BodyAction                
                      { $$ = nn(',', ZC, ZS, $1, $3); }   
        | OUTPUT SignalName TO BufferName          
                      { $$ = nn(OUTPUT, ZC, ZS, $2, $4);  }
        | OUTPUT SignalName '(' ExpressionArgList ')' TO BufferName
                      { 
			/* BUG !!! any expression */
			/* label (replace) any's in output commands */
			replace_any($4);
                        $2->lft = $4;
			$$ = nn(OUTPUT, ZC, ZS, $2, $7);
		      }
        | SET Expression ASGN Expression           
                      { $$ = nn(SET, ZC, ZS, $2, $4); }
        | RESET Expression                         
                      { $$ = nn(RESET, ZC, ZS, $2, ZN); }
        | Expression ASGN Expression               
                      { $$ = nn(ASGN, ZC, ZS, $1, $3); }
        | Expression                               
                      { $$ = nn('*', ZC, ZS, $1, ZN); }
        | '{' BodyAction '}'                       
                      { $$ = nn('{', ZC, ZS, $2, ZN); }
        ;




/*
----------------------------------------------------------------------------
Expression Syntax
----------------------------------------------------------------------------
*/

Expression: 
	  Integer                                  { $$ = $1; }
        | Real                                     { $$ = $1; }
        | ANY                                      { $$ = $1; }
/* | SELF                                     { $$ = $1; }
        | SENDER                                   { $$ = $1; } */
        | Name                                     { $$ = $1; }
        | Name '(' ExpressionArgOptList ')'        { $1->lft = $3; $$ = $1;} 
        | Expression '.' Name                 { $$ = nn('.', ZC, ZS, $1, $3); }
        | Expression '[' Expression ']'       { $$ = nn('[', ZC, ZS, $1, $3); }
        | ACTIVE Expression                { $$ = nn(ACTIVE, ZC, ZS, $2, ZN); }
        | PRED Expression                    { $$ = nn(PRED, ZC, ZS, $2, ZN); }
        | SUCC Expression                    { $$ = nn(SUCC, ZC, ZS, $2, ZN); }
        | POS Expression                     { $$ = nn(POS, ZC, ZS, $2, ZN); }
        | NOT Expression                     { $$ = nn(NOT, ZC, ZS, $2, ZN);}
        | '+' Expression %prec UMIN          { $$ = nn(UMIN, ZC, ZS, $1, $2); }
        | '-' Expression %prec UMIN          { $$ = nn(UMIN, ZC, ZS, $1, $2); }
        | Expression '*' Expression      { $$ = nn('*', $2->txt, ZS, $1, $3); }
        | Expression '/' Expression      { $$ = nn('/', $2->txt, ZS, $1, $3); }
        | Expression '%' Expression      { $$ = nn('%', $2->txt, ZS, $1, $3); }
        | Expression '+' Expression      { $$ = nn('+', $2->txt, ZS, $1, $3); }
        | Expression '-' Expression      { $$ = nn('-', $2->txt, ZS, $1, $3); }
        | Expression '<' Expression      { $$ = nn('<', $2->txt, ZS, $1, $3); }
        | Expression '>' Expression      { $$ = nn('>', $2->txt, ZS, $1, $3); }
        | Expression LE Expression       { $$ = nn(LE, $2->txt, ZS, $1, $3); }
        | Expression GE Expression       { $$ = nn(GE, $2->txt, ZS, $1, $3); }
        | Expression '=' Expression          { $$ = nn('=', ZC, ZS, $1, $3); }
        | Expression NE Expression           { $$ = nn(NE, ZC, ZS, $1, $3); } 
        | Expression AND Expression          { $$ = nn(AND, ZC, ZS, $1, $3); }
        | Expression OR Expression           { $$ = nn(OR, ZC, ZS, $1, $3); }
        | Expression '?' Expression ':' Expression 
                                             {
					       $$ = nn(':', $4->txt, ZS, $3, $5);
					       $$ = nn('?', ZC, ZS, $1, $$);
					     }
        | '(' Expression ')'                       
                                             { 
					       $$ = nn('(', ZC, ZS, $2, ZN);
					     }
        ;

ExpressionArgOptList:
          ExpressionArgList                        { $$ = $1; }
        | /* empty */                              { $$ = ZN; }
        ;

ExpressionArgList:
          ExpressionArgList ',' Expression         { $$ = insertl($1,$3); }
        | Expression                               { $$ = $1; }
        ;

Offert: 
          '!' Expression                     { }
        | '?' Expression %prec '!'           { }
        ;

OffertOptSet: 
          OffertOptSet Offert                      { $1->rgt = $2; $$ = $1; }
        | /* empty */                              { $$ = ZN; }
        ;




/*
----------------------------------------------------------------------------
Options
----------------------------------------------------------------------------
*/

OptionSet: 
          OptionSet Option                         { $$ = insertl($1,$2); }
        | /* empty */                              { $$ = ZN; }
        ;

Option: 
          ':' Name                                 { $$ = $2; }
        | ':' Name Integer                         { $2->lft = $3; $$ = $2; } /* !? */
        ;




/*
----------------------------------------------------------------------------
Miscelaneuous
----------------------------------------------------------------------------
*/

Name: 
          NAME                    { /* BUG !!! check for collision 
				       with Promela keywords */
                                    if (strcmp($1->txt, "init")==0)
                                      strcpy($1->txt, "__init");
                                    $$ = $1; }
        ;

String:   
          STRING                 { $$ = $1; }
        ;

Integer: 
          INTEGER                { $$ = $1; }
        ;

Real:                     
          FLOAT                  { $$ = $1; }
        ;

%%








