#include "if2pml.h"
#include "y.tab.h"

extern Node *syst;

int MaxSlotSize, BuffSize = 4; /* maximal slot and buffer length */

FILE *out;

#define PNode(x)       ((Node *) (x))
#define PSyStruct(x)   ((SyStruct *) (x))
#define PSigStruct(x)   ((SigStruct *) (x))
#define PBuffStruct(x) ((BuffStruct *) (x))
#define PProcStruct(x) ((ProcStruct *) (x))
#define PVarStruct(x)  ((VarStruct *) (x))
#define PStateStruct(x)((StateStruct *) (x))
#define PTransStruct(x)((TransStruct *) (x))
#define PInputStruct(x)((InputStruct *) (x))
#define PFilterActionStruct(x)((FilterActionStruct *) (x))

/* char in_body; */ /* !? */ /* indicator for expire handling */;



void translate(char OutFile[])
{

  if ( (out = fopen(OutFile, "w")) == NULL){
    sprintf(IfParserMessage, "can't open an ouput file");
    IF_ERROR(IfParserMessage);
  }

  headings();
  types();
  signals();
  buffers();
  processes();
}


void headings()
{
  fprintf(out, "/* System %s */\n\n", syst->txt);
  fprintf(out, "#include \"dtimesdl.h\"\n\n");
  fprintf(out, "#define _any 0\n\n");

  

} /* headings() */


void types()
{
  Node *p, *q, *r; 
  int val;

#ifdef DEBUG
  printf("types\n");
#endif

  if (!(PSyStruct(syst->nstr)->types)) return ; /* no types */

  fprintf(out, "\n/* Types */\n\n");

  /* walk down the type list */
  for(p = PSyStruct(syst->nstr)->types; p; p = p->rgt)
    {
      switch(p->ntyp)
	{
	case ENUM: 
#ifdef DEBUG
          printf("J'etais ici - ENUM\n");
#endif
	  fprintf(out, "/* enum */\n#define %s byte\n", p->txt);
	  /* item names */
	  for (val = 0, q = p->lft; q; val++, q = q->rgt)
	    fprintf(out, "#define %s %d\n", q->txt, val);
	  fprintf(out, "/* end enum */\n\n");
	  break;
	case RANGE:
#ifdef DEBUG
          printf("J'etais ici - RANGE\n"); 
#endif
	  fprintf(out, "/* range */\n#define %s int\n\n", p->txt);
	  break;
	case RECORD:
#ifdef DEBUG
          /* printf("J'etais ici - RECORD\n"); */
#endif
	  fprintf(out, "typedef %s {\n", p->txt);
	  /* record fields */
	  vars(p->lft, ZN);
#ifdef DEBUG
           printf("J'etais ici - RECORD\n");
#endif
	  fprintf(out, "}\n\n");
	  break;
	case ARRAY:
#ifdef DEBUG
          /* printf("J'etais ici - ARRAY\n"); */
#endif
          if (p->lft->lft->ntyp == '-')
	    yyerror("negative ints or enums as array indices");
	  /* fprintf(out, "%s %s[%s];\n\n", p->lft->rgt->txt, p->txt, 
		  p->lft->lft->rgt->txt); */
          if (!(p->lft->lft->lft)) yyerror("Tuka e greskata!\n");
          fprintf(out, "%s %s[%s];\n\n", p->lft->rgt->txt, p->txt, 
		  p->lft->lft->rgt->txt /* p->lft->lft->rgt->txt */);  
	  break;
	case ABSTRACT:
	  fprintf(out, "/* warning: abstract %s ignored! */\n\n", p->txt);
	}
	
    }; 
} /* types() */


void signals()
{
  
  Node *p; Node *q; int nr_arg = 0;

#ifdef DEBUG  
  printf("Signals\n");
#endif

  if (!(PSyStruct(syst->nstr)->signals)) return ; /* no signals */

  fprintf(out, "\n/* Signals */\n\n"); 

  fprintf(out, "mtype = {\n"); /* open mtype list */
  
  /* walk down the signal list */
  for(p = PSyStruct(syst->nstr)->signals; p; p = p->rgt)
    {
      

      fprintf(out, "%s%s\n", p->txt, (p->rgt?",":""));
      /* count the number of arguments */
      if (p->lft)
	for (nr_arg = 0, q = p->lft; q;
	     q = q->rgt /* , nr_arg += type_size(q) */)
	  {
	 
	  nr_arg += type_size(q);
	  };
      

      PSigStruct(p->nstr)->eff_nr_arg = nr_arg;
      if (nr_arg > MaxSlotSize) MaxSlotSize = nr_arg;
    };

  fprintf(out, "};\n"); /* close  mtype list */
} /* signals() */



int type_size(Node *n)
{
  Node *p, *q, *r;
  int size, fields;

#ifdef DEBUG
  printf("type_size\n");  
#endif

  if (n && n->ntyp != NAME)
    return 1;
      /* user defined type */
      /* find the type */
  if (!(PSyStruct(syst->nstr)->types)) 
    yyerror("wrong type identifier");

  for (p = PSyStruct(syst->nstr)->types; p; p = p->rgt)
    {
    if (strcmp(n->txt, p->txt) == 0)
	/* compute the size */
	switch (p->ntyp)
	{
	case ENUM:
	case RANGE: 
	  return 1;
	case ARRAY: 
	  /* Possible BUG !!! if an array element a[n] is send */
	  return atoi(p->lft->lft->rgt->txt);
	  /* right range limit */ 
	case RECORD:
	  /* for each record field */
	  for (size = 0, q = n->lft; q; q = q->rgt)
	    {
	      /* find the type - if there is a list of fields with a common 
		 type they are listed as left children of the first var */
	      for (r = q, fields = 1; !(PVarStruct(r->nstr)->vtyp); 
		   r = r->lft, fields++);
	      size += fields*type_size(r);
	    };
	  return size;
	case ABSTRACT: 
	  /* yyerror("ADT as signal argument"); BUG !!! */ 
	  return 1;
	default: 
	  yyerror("Unknown data type");
	}; /* switch */
    
   

    }
} /* type size() */



Node *find_var_type(Node *n, Node *proc)
{
  Node *p, *q;
  int found = 0;

#ifdef DEBUG
  printf("find_var_type\n");
  if (n->ntyp == NAME) printf("varname1: %s", n->txt); 
#endif
  
  for (p = PProcStruct(proc->nstr)->vars; p; p = p->rgt)
    {
    /* if there is a list of vars with a common 
       type they are listed as left children of the first var */

      if (  PVarStruct(p->nstr)->vtyp &&
            ((n->ntyp == NAME) && (strcmp(p->txt, n->txt) == 0) ||
	    (n->ntyp == '[') && (strcmp(p->txt, n->lft->txt) == 0))
         )			  return PVarStruct(p->nstr)->vtyp;


      for (q = p; !(PVarStruct(q->nstr)->vtyp); q = q->lft)
      {
#ifdef DEBUG
   printf("varname:%s", q->txt);
#endif        
	if ((n->ntyp == NAME) && (strcmp(q->txt, n->txt) == 0) ||
	    (n->ntyp == '[') && (strcmp(q->txt, n->lft->txt) == 0))			  found = 1;	
      };
    
     if (  PVarStruct(q->nstr)->vtyp &&
            ((n->ntyp == NAME) && (strcmp(q->txt, n->txt) == 0) ||
	    (n->ntyp == '[') && (strcmp(q->txt, n->lft->txt) == 0))
         )			  found = 1;

    if (found)
	return PVarStruct(q->nstr)->vtyp;
    };
  yyerror("undeclared variable");
} /* find_var_type() */


int enum_const(Node *n)
{
  Node *p;

#ifdef DEBUG
  printf("enum_constant\n");
#endif

  if (!(PSyStruct(syst->nstr)->types)) 
    /* no user defined types */
    return 0;
  
  /* check all enum types */
  for (p = PSyStruct(syst->nstr)->types; p; p = p->rgt)
    if (p->ntyp == ENUM)
	if (find_symbol(n, p->lft)) return 1;
  return 0; /* not an enum constant */
} /* enum_const() */


Node *find_symbol(Node *symbol, Node *list)
{
  Node *r;

#ifdef DEBUG
  printf("find_symbol - entry\n");  
#endif  


  for (r = list; r && (strcmp(r->txt, symbol->txt) != 0); r = r->rgt);

#ifdef DEBUG
   printf("find_symbol -exit\n");
#endif

  return r;



} /* find_symbol */



void buffers()
{

  int i;
  Node *p, *q, *r;

#ifdef DEBUG
  printf("buffers\n");
#endif

  if (!(PSyStruct(syst->nstr)->buffers)) return; /* no buffers */ 

  fprintf(out, "\n/* Buffers */\n"); 

  fprintf(out, "\n#define BUFFSIZE %d\n\n", BuffSize);

  for(p = PSyStruct(syst->nstr)->buffers; p; p = p->rgt)
    {
      fprintf(out, "chan %s = [BUFFSIZE] of {mtype", p->txt);
      fprintf(out, ", byte"); /* potential BUG if no_pid !!! */

      /* compute the slot size */
      /* if (!(PBuffStruct(p->nstr)->signals)) 
	yyerror("missing buffer signal list"); */ 
      if (strcmp(PBuffStruct(p->nstr)->signals->txt, "*") == 0)
	/* all types of signals */
	PBuffStruct(p->nstr)->slotsize = MaxSlotSize;
      else  /* walk the buffer's signal list */
      for(q = PBuffStruct(p->nstr)->signals; q; q = q->rgt)
	{
	  /* find the signal in the system signal list */
	  
	  r = find_symbol(q, PSyStruct(syst->nstr)->signals);
	  if (!r) yyerror("buffers: undeclared signal");

	  if (PSigStruct(r->nstr)->eff_nr_arg > PBuffStruct(p->nstr)->slotsize)
	     PBuffStruct(p->nstr)->slotsize = PSigStruct(r->nstr)->eff_nr_arg;
        }

      /* complete the buffer slot */
      for (i=1; i<PBuffStruct(p->nstr)->slotsize; i++)
      /* for (i=1; i<MaxSlotSize; i++) */
	fprintf(out, ", int");
      fprintf(out, "};\n");
    };

} /* buffers() */


char *BuffType(Node *signals)
{
  return "mtype";
} /* BuffType() */


void processes()
{
  Node *p, *q, *r, *s;

  
  int nr_args, /* the "effective" nr of signal arguments */ 
      nr_trans = 0; /* nr of transitions */
  
#ifdef DEBUG
  printf("processes\n");
#endif

  if (!(PSyStruct(syst->nstr)->processes)) return; /* !? */

  fprintf(out, "\n/* Processes */\n\n");

  for(p = PSyStruct(syst->nstr)->processes; p; p = p->rgt)
  {
    /* proc name */
    /* fprintf(out, "\nproctype %s_proc(%s)\n{\n", p->txt, 
                 (PProcStruct(p->nstr)->instances ? "short _inst" : "" )); */

    fprintf(out, "\nproctype %s()\n{\n", p->txt);

    /* vars */
    if (PProcStruct(p->nstr)->vars)
    for (q = PProcStruct(p->nstr)->vars; q; q = q->rgt)
      {
	/* find the type - if there is a list of vars with a common 
           type they are listed as left children of the first var */
	for (r = q; !(PVarStruct(r->nstr)->vtyp); r = r->lft);

	switch (PVarStruct(r->nstr)->vtyp->ntyp)
	  {
	  case NAME: fprintf(out, "%s\n", PVarStruct(r->nstr)->vtyp->txt); break;
	  case BOOL: fprintf(out, "bool\n"); break;
	  case PID: fprintf(out, "byte\n"); break;
	  case INT:  fprintf(out, "int\n"); break;
          case REAL: fprintf(out, "int\n"); break;
	  case TIMER: fprintf(out, "timer\n"); break;
	  default: IF_ERROR("Unknown type");
	  };


	/* var names */
	for (r = q; r ;r = r->lft)
	  {
	    fprintf(out, "\t%s", r->txt);
	    /* defalt value (initialization) (NYI!!) */
	    fprintf(out, "%s", PVarStruct(r->nstr)->dflt ? " = " : "");
	    if (PVarStruct(r->nstr)->dflt)
	      PrintExpr(PVarStruct(r->nstr)->dflt, p);
	    fprintf(out, "%s", r->lft ? ",\n" : ";\n");
	  };
      };

    /* for each state */
    for (q = PProcStruct(p->nstr)->states; q; q = q->rgt)
      {  
	
	/* if stable */
	if (PStateStruct(q->nstr)->options &&
	    strcmp(PStateStruct(q->nstr)->options->txt, "nostable")==0)
	    fprintf(out, "\natomic{\nskip;\n%s:\n", q->txt);
        else  
	    fprintf(out, "\n%s:\natomic{\n", q->txt);
	
	fprintf(out, "if\n");

      /* find all transitions originationg in this state */
	for (r = PProcStruct(p->nstr)->trans; r; r = r->rgt)
	  {
#ifdef DEBUG
	    printf("%s=%s\n", q->txt,PTransStruct(r->nstr)->source->txt );
#endif
	    if (strcmp(PTransStruct(r->nstr)->source->txt, q->txt)==0)
	      {
		nr_trans++;
		fprintf(out, ":: ");
		/* guard */
		if (PTransStruct(r->nstr)->guard)
		  {
		    PrintExpr(PTransStruct(r->nstr)->guard,p);
		    fprintf(out, "->\n\t");
		  };

		/* input */
		if (PTransStruct(r->nstr)->input)
		  {
		    /* channel */
		    fprintf(out, "%s?",
                    PInputStruct(PTransStruct(r->nstr)->input->nstr)->
                                 frombuff->txt);
		   
		    /* signal */
		    fprintf(out, "%s", PInputStruct(PTransStruct(r->nstr)
                                       ->input->nstr)->signal->txt);
		    if (PInputStruct(PTransStruct(r->nstr)->input->nstr)
			               ->sigexpr)
		      {
			/* attributes (parameters) */
			fprintf(out, "(");
			for(nr_args = 0, s = PInputStruct(PTransStruct(r->nstr)
				->input->nstr)->sigexpr;s; s = s->rgt)
			  {
			    if ((s->ntyp != NAME) || (s->ntyp != '['))
			      /* not a user def var  or array */
			      nr_args++;
			    else
			      nr_args += type_size(find_var_type(s,p));
			    PrintExpr(s,p); 
			    fprintf(out, "%s", (s->rgt ? "," : "")); 
			  };

			/* padding of the "missing" parameters */
			/* find the slot size */
			s = find_symbol(PInputStruct(PTransStruct(r->nstr)->
						 input->nstr)->frombuff,
				    PSyStruct(syst->nstr)->buffers);
			if (!s) yyerror("processes: undeclared buffer");
			for (; nr_args < PBuffStruct(s->nstr)->slotsize;
			     nr_args++)
			  fprintf(out, ",_");
			fprintf(out, ")");
		      };
		    fprintf(out, "->\n\t");
		  }; /* end of input */
	      
	       /* body */
	       if (PTransStruct(r->nstr)->body)
		 {
		   body(PTransStruct(r->nstr)->body,p);
		   fprintf(out, "->\n\t");
		 };

	       /* target */
	       fprintf(out, "goto %s;\n", 
		       (PTransStruct(r->nstr)->target->ntyp == '-'?
			q->txt :
			PTransStruct(r->nstr)->target->txt));  
	      }; /* rof one transition */    
	  }; /* rof transitions */
	
	/* discards */
	if (PStateStruct(q->nstr)->discards)
	  {
	    nr_trans++;
            discards(PStateStruct(q->nstr)->discards, q, p);
	  };

	/* close state */
	if (!nr_trans) fprintf(out, ":: false\n");
	nr_trans = 0;
	fprintf(out, "fi;\n}\n");
      }; /* rof state */
    fprintf(out, "};\n\n"); /*end of process */
    }; /* rof processes */

  /* run the processes from init */
  fprintf(out, "init{\n\tatomic{\n");
  for(p = PSyStruct(syst->nstr)->processes; p; p = p->rgt)
    fprintf(out, "\t\trun %s();\n", p->txt);
  fprintf(out, "\t}\n}");
	    
} /* processes() */

  
void vars( Node *p, Node *pr)
{
  Node *q, *r;
  
#ifdef DEBUG
  printf("vars\n");
#endif

  for (q = p; q; q = q->rgt)
      {
	/* find the type - if there is a list of vars with a common 
           type they are listed as left children of the first var */
	for (r = q; !(PVarStruct(r->nstr)->vtyp); r = r->lft);

	switch (PVarStruct(r->nstr)->vtyp->ntyp)
	  {
	  case NAME: fprintf(out, "%s\n", PVarStruct(r->nstr)->vtyp->txt); break;
	  case BOOL: fprintf(out, "bool\n"); break;
	  case PID: fprintf(out, "byte\n"); break;
	  case INT:  fprintf(out, "int\n"); break;
          case REAL: fprintf(out, "int\n"); break;
	  case TIMER: fprintf(out, "timer\n"); break;
	  default: IF_ERROR("Unknown type");
	  };


	/* var names */
	for (r = q; r ;r = r->lft)
	  {
	    fprintf(out, "\t%s", r->txt);
	    /* defalt value (initialization) (NYI!!) */
	    fprintf(out, "%s", PVarStruct(r->nstr)->dflt ? " = " : "");
	    if (PVarStruct(r->nstr)->dflt)
	      PrintExpr(PVarStruct(r->nstr)->dflt, pr);
	    fprintf(out, "%s", r->lft ? ",\n" : ";\n");
	  };

      };
} /* vars() */

void PrintExpr(Node *n, Node *pr /* process */)
{
  /*  fprintf(out, "Expression"); */

#ifdef DEBUG
  printf("PrintExpr\n");
#endif

  switch (n->ntyp)
    {
    case ANY:    if (strcmp(n->txt,"any_snd") == 0) 
                    fprintf(out, "_any");
                 else
		    fprintf(out, "_");
                 break;
    case INTEGER: 
    case NAME:   
    case UMIN:   fprintf(out,"%s",  n->txt); 
                 break;
		 /*  case SENDER: fprintf(out, "_sender");
                 break; */
    case FLOAT:  fprintf(out, "%d /* warning: converted float */",
			 atoi(n->txt));  
                 /* BUG !!! float converted to integer */
                 break;
    case '[':    PrintExpr(n->lft, pr); fprintf(out, "["); 
                 PrintExpr(n->rgt, pr); fprintf(out, "]");
		 break;
		 
    case '.':    PrintExpr(n->lft, pr); fprintf(out, "."); 
                 PrintExpr(n->rgt, pr); 
                 break;
                 
    case NOT:    fprintf(out, "!"); PrintExpr(n->lft, pr);
                 break;
    case '*':
    case '/':
    case '%':
    case '+':
    case '-':
    case '<':
    case '>':
    case LE:
    case GE:
    case ':':   PrintExpr(n->lft, pr); fprintf(out, "%s", n->txt, pr); 
                PrintExpr(n->rgt, pr);
		break;
    case '=':   if (is_timer_name(n->lft->txt, pr))
                  {
		    if (strcmp(n->rgt->txt, "0") == 0)
                      {
		         fprintf(out, "expire(%s)", n->lft->txt);
		         break;
		      }
                    else IF_ERROR("timer compared to nonzero value");
		  };
                if (is_timer_name(n->rgt->txt, pr))
		  {
		   if (strcmp(n->lft->txt, "0") == 0)
                      {
		         fprintf(out, "expire(%s)", n->rgt->txt);
			 break;
		      }
                   else IF_ERROR("timer comapred to nonzero value");
		  };
                /* no timers */
                PrintExpr(n->lft, pr); fprintf(out, "==");
                PrintExpr(n->rgt, pr);  
		break;
    case NE:    PrintExpr(n->lft, pr); fprintf(out, "!="); 
                PrintExpr(n->rgt, pr);
		break;		
    case AND:   PrintExpr(n->lft, pr); fprintf(out, " && "); 
                PrintExpr(n->rgt, pr);
		break;
    case OR:    PrintExpr(n->lft, pr); fprintf(out, "||"); 
                PrintExpr(n->rgt, pr);
		break; 
    case '?':   PrintExpr(n->lft, pr); fprintf(out, "->"); 
                PrintExpr(n->rgt, pr);
		break;
    case '(':   fprintf(out, "("); PrintExpr(n->lft, pr);
                fprintf(out, ")");
		break;

    default: IF_ERROR("wrong expression type");
    }
} /* PrintExpr() */ 



int is_timer_name(char *name, Node *pr)
{
  Node *p, *q, *r;
 
#ifdef DEBUG
  printf("is_timer_name\n");
#endif

  /* search process vars */
  for(p = PProcStruct(pr->nstr)->vars; p; p = p->rgt)
      if (strcmp(name,p->txt) == 0)         /* var found */
	{
	  /* find the type */
	  for(q = p; !(PVarStruct(q->nstr)->vtyp) && q; q = q->lft);
	  if (!(PVarStruct(q->nstr)->vtyp))
	    IF_ERROR("variable without type");
	  if (PVarStruct(q->nstr)->vtyp->ntyp == TIMER)
	    return 1;
	}
      else 
	if (p->lft)/* several variables of same type */
	  {
	    for(q = p->lft; q; q = q->lft)
	    if (strcmp(name,q->txt) == 0)         /* var found */
	      {
		/* find the type */
		for(r = q; !(PVarStruct(r->nstr)->vtyp); r = r->lft); 
		if (strcmp(PVarStruct(r->nstr)->vtyp->txt, "TIMER") == 0)
		  return 1;
	      }
	  };  
	 
  /* search system vars */ /* ?! */ /* not needed? */
  return 0;
} /* is_timer_name() */



void body(Node *n, Node *pr /* process */) 
{
  Node *s;
  int nr_arg;

#ifdef DEBUG
  printf("body\n");
#endif

  switch (n->ntyp)
    {
      /* case '|': */
    case ',':
      body(n->lft, pr); fprintf(out, ";\n\t"); body(n->rgt, pr);
      break;
    case OUTPUT:
      fprintf(out, "%s!",n->rgt->txt); 
      fprintf(out, "%s",n->lft->txt);
      /* arguments */
      if (n->lft->lft)
	{
	  fprintf(out, "(");
	  for(nr_arg = 0, s = n->lft->lft; s; s = s->rgt)
	    {
	      if ((s->ntyp != NAME) || (strcmp(s->txt, pr->txt) == 0) ||
		  enum_const(s)) 
		nr_arg++;
	      else
		nr_arg += type_size(find_var_type(s,pr));
	      if (strcmp(s->txt, pr->txt) == 0)
		fprintf(out, "_pid");
	      else	      
		fprintf(out, "%s", s->txt);
	      fprintf(out, "%s", s->rgt ? ", " : "");
	    };
	};
      /* dummy arguments */
      /* find max slot size */
      s = find_symbol(n->rgt, PSyStruct(syst->nstr)->buffers);

      /* Here is the problem */
      if (!s) yyerror("Unknown buffer");

      for (;nr_arg < PBuffStruct(s->nstr)->slotsize; ++nr_arg)
	  fprintf(out, ",0");
      fprintf(out, ")"); /* potential BUG !!! if there is a no_pid switch */
      break;
    case SET:
      fprintf(out, "set("); PrintExpr(n->lft,pr);
      fprintf(out, ","); PrintExpr(n->rgt,pr);
      fprintf(out, ")");
      break;
    case RESET:
      fprintf(out, "reset("); PrintExpr(n->lft,pr); 
      fprintf(out, ")");
      break;  
    case ASGN:
      PrintExpr(n->lft,pr); fprintf(out, " = ");
      /* check if the process name is being assigned */
      if (strcmp(n->rgt->txt, pr->txt)==0) strcpy(n->rgt->txt, "_pid");
      PrintExpr(n->rgt,pr);	      
      break;
    case '{':
      fprintf(out, "{"); body(n->lft, pr); fprintf(out, "}");
      break;
    case '*': 
      PrintExpr(n->lft,pr);
      break;
    default: IF_ERROR("Undefined body action");
    }; /* case */
}/* body() */


void discards(Node *n, Node *state, Node *pr)
{
  Node *p, *q, *r;
  FilterActionStruct *pfa;
  int i;

#ifdef DEBUG
  printf("discards\n");
#endif

  /* for all filter actions (buffers) */
  for(p = n; p; p = p->rgt)
    { 
      pfa = PFilterActionStruct(n->nstr);
      if (strcmp(pfa->signals->txt, "*") == 0)
	{
	  /* find all signals for this buffer */
	  /* BUG !!! */
	  yyerror("discard \"*\" not implemented");
	  /* loop; */
	};
      /* for all signals */
      for(q = pfa->signals; q; q = q->rgt)
	{
	  /* make a transition */
	  fprintf(out, ":: ");
	  /* add an optional guard */
	  if (pfa->guard)
	  {
	    PrintExpr(pfa->guard,pr); fprintf(out, "&&");
	  };
	  fprintf(out, "%s?", pfa->buffer->txt);
	  fprintf(out, "%s%s",(pfa->guard ? "[" : ""), q->txt);
		  /* arguments (trouble) */ /* !? */ /* BUG !!! */
	  /* find max slot size */
	  r = find_symbol(pfa->buffer, PSyStruct(syst->nstr)->buffers);
	  if (!r) yyerror("discards: undeclared buffer");
	  for(i=0; i < PBuffStruct(r->nstr)->slotsize; i++)
	      fprintf(out, ",_");
		  /* if sig_has_arguments() */
	  fprintf(out,"%s->\n\t", (pfa->guard ? "]" : ""));
	  fprintf(out, "goto %s; /* discard */\n", state->txt);   
	}; /* signals */
    }; /* actions */
} /* discards() */





